// material.cpp

// includes
#include <cstring>
#include <cstdlib>
#include "option.h"
#include "hash.h"
#include "init.h"
#include "eval.h"
#include "uci.h"
#include "material.h"
#include "search.h"

// types
typedef material_info_t entry_t;

struct material_t
    {
    entry_t *table;
    uint32 size;
    uint32 mask;
    uint32 used;
    sint64 read_nb;
    sint64 read_hit;
    sint64 write_nb;
    sint64 write_collision;
    };
// variables
static material_t material[16][1];

// prototypes
static void material_comp_info(material_info_t *info, const board_t *board);

// functions

// material_init()
void material_init()
    {
    int thread_id;

    // material table
    for (thread_id = 0; thread_id < number_threads; thread_id++)
        {
        material[thread_id]->size = 0;
        material[thread_id]->mask = 0;
        material[thread_id]->table = NULL;
        }
    }

// material_alloc()
void material_alloc()
    {
    int thread_id;

    if(use_material_table)
        {
        for (thread_id = 0; thread_id < number_threads; thread_id++)
            {
            material[thread_id]->size = material_table_size;
            material[thread_id]->mask = material_table_size - 1;
            material[thread_id]->table = (entry_t *)malloc(material[thread_id]->size * sizeof(entry_t));
            material_clear(thread_id);
            }
        }
    }

// material_clear()
void material_clear(int thread_id)
    {
    if(material[thread_id]->table != NULL)
        {
        memset(material[thread_id]->table, 0, material[thread_id]->size * sizeof(entry_t));
        }
    material[thread_id]->used = 0;
    material[thread_id]->read_nb = 0;
    material[thread_id]->read_hit = 0;
    material[thread_id]->write_nb = 0;
    material[thread_id]->write_collision = 0;
    }

// material_get_info()
void material_get_info(material_info_t *info, const board_t *board, int thread_id)
    {
    uint64 key;
    entry_t *entry;

    // probe
    if(use_material_table)
        {
        material[thread_id]->read_nb++;
        key = board->material_key;
        entry = &material[thread_id]->table[KEY_INDEX(key) & material[thread_id]->mask];

        if(entry->lock == KEY_LOCK(key))
            {
            // found
            material[thread_id]->read_hit++;
            *info = *entry;
            return;
            }
        }
    // calculation
    material_comp_info(info, board);

    // store
    if(use_material_table)
        {
        material[thread_id]->write_nb++;

        if(entry->lock == 0)
            {
            material[thread_id]->used++;
            }
        else
            {
            material[thread_id]->write_collision++;
            }
        *entry = *info;
        entry->lock = KEY_LOCK(key);
        }
    }

// material_comp_info()
static void material_comp_info(material_info_t *info, const board_t *board)
    {
    int wp, wn, wb, wr, wq;
    int bp, bn, bb, br, bq;
    int wt, bt;
    int wm, bm;
    int tp;
    int np;
    int tm;
    int color;
    int recog;
    int flags;
    int cflags[2];
    int mul[2];
    int phase;
    int opening, endgame;
    int owf, obf, ewf, ebf;
    // init
    wp = board->number[0];
    wn = board->number[2];
    wb = board->number[4];
    wr = board->number[6];
    wq = board->number[8];
    bp = board->number[1];
    bn = board->number[3];
    bb = board->number[5];
    br = board->number[7];
    bq = board->number[9];
    wt = wq + wr + wb + wn + wp; // no king
    bt = bq + br + bb + bn + bp; // no king
    wm = wb + wn;
    bm = bb + bn;
    tp = wp + bp;
    np = wq + wr + wm + bq + br + bm;
    tm = wm + bm;
    // recogniser
    recog = MAT_NONE;

    if(false) { }
    else if(wt == 0 && bt == 0)
        {
        recog = MAT_KK;
        }
    else if(wt == 1 && bt == 0)
        {
        if(wb == 1)
            recog = MAT_KBK;

        if(wn == 1)
            recog = MAT_KNK;

        if(wp == 1)
            recog = MAT_KPK;
        }
    else if(wt == 0 && bt == 1)
        {
        if(bb == 1)
            recog = MAT_KKB;

        if(bn == 1)
            recog = MAT_KKN;

        if(bp == 1)
            recog = MAT_KKP;
        }
    else if(wt == 1 && bt == 1)
        {
        if(wq == 1 && bq == 1)
            recog = MAT_KQKQ;

        if(wq == 1 && bp == 1)
            recog = MAT_KQKP;

        if(wp == 1 && bq == 1)
            recog = MAT_KPKQ;

        if(wr == 1 && br == 1)
            recog = MAT_KRKR;

        if(wr == 1 && bp == 1)
            recog = MAT_KRKP;

        if(wp == 1 && br == 1)
            recog = MAT_KPKR;

        if(wb == 1 && bb == 1)
            recog = MAT_KBKB;

        if(wb == 1 && bp == 1)
            recog = MAT_KBKP;

        if(wp == 1 && bb == 1)
            recog = MAT_KPKB;

        if(wn == 1 && bn == 1)
            recog = MAT_KNKN;

        if(wn == 1 && bp == 1)
            recog = MAT_KNKP;

        if(wp == 1 && bn == 1)
            recog = MAT_KPKN;
        }
    else if(wt == 2 && bt == 0)
        {
        if(wb == 1 && wp == 1)
            recog = MAT_KBPK;

        if(wn == 1 && wp == 1)
            recog = MAT_KNPK;
        }
    else if(wt == 0 && bt == 2)
        {
        if(bb == 1 && bp == 1)
            recog = MAT_KKBP;

        if(bn == 1 && bp == 1)
            recog = MAT_KKNP;
        }
    else if(wt == 2 && bt == 1)
        {
        if(wr == 1 && wp == 1 && br == 1)
            recog = MAT_KRPKR;

        if(wb == 1 && wp == 1 && bb == 1)
            recog = MAT_KBPKB;
        }
    else if(wt == 1 && bt == 2)
        {
        if(wr == 1 && br == 1 && bp == 1)
            recog = MAT_KRKRP;

        if(wb == 1 && bb == 1 && bp == 1)
            recog = MAT_KBKBP;
        }
    // draw node (exact-draw recogniser)
    flags = 0;

    for (color = 0; color < 2; color++)
        cflags[color] = 0;

    if(wq + wr + wp == 0 && bq + br + bp == 0)
        {                // no major piece or pawn
        if(wm + bm <= 1 // at most one minor => KK, KBK or KNK
        || recog == MAT_KBKB)
            {
            flags |= (1 << 0);
            }
        }
    else if(recog == MAT_KPK || recog == MAT_KKP || recog == MAT_KBPK || recog == MAT_KKBP)
        {
        flags |= (1 << 0);
        }

    // bishop endgame
    if(wq + wr + wn == 0 && bq + br + bn == 0)
        { // only bishops
        if(wb == 1 && bb == 1)
            {
            if(wp - bp >= -2 && wp - bp <= +2)
                { // pawn diff <= 2
                flags |= (1 << 1);
                }
            }
        }

    // multipliers
    for (color = 0; color < 2; color++)
        mul[color] = 16; // 1
    // white multiplier
    if(wp == 0)
        { // white has no pawns
        int w_maj = wq * 2 + wr;
        int w_min = wb + wn;
        int w_tot = w_maj * 2 + w_min;
        int b_maj = bq * 2 + br;
        int b_min = bb + bn;
        int b_tot = b_maj * 2 + b_min;

        if(false) { }
        else if(w_tot == 1)
            {
            // KBK* or KNK*, always insufficient
            mul[0] = 0;
            }
        else if(w_tot == 2 && wn == 2)
            {
            // KNNK*, usually insufficient
            if(b_tot != 0 || bp == 0)
                {
                mul[0] = 0;
                }
            else
                {           // KNNKP+, might not be draw
                mul[0] = 1; // 1/16
                }
            }
        else if(w_tot == 2 && wb == 2 && b_tot == 1 && bn == 1)
            {
            // KBBKN*, barely drawish (not at all?)
            mul[0] = 8; // 1/2
            }
        else if(w_tot - b_tot <= 1 && w_maj <= 2)
            {
            // no more than 1 minor up, drawish
            mul[0] = 2; // 1/8
            }
        }
    else if(wp == 1)
        { // white has one pawn
        int w_maj = wq * 2 + wr;
        int w_min = wb + wn;
        int w_tot = w_maj * 2 + w_min;
        int b_maj = bq * 2 + br;
        int b_min = bb + bn;
        int b_tot = b_maj * 2 + b_min;

        if(false) { }
        else if(b_min != 0)
            {
            // assume black sacrifices a minor against the lone pawn
            b_min--;
            b_tot--;

            if(false) { }
            else if(w_tot == 1)
                {
                // KBK* or KNK*, always insufficient
                mul[0] = 4; // 1/4
                }
            else if(w_tot == 2 && wn == 2)
                {
                // KNNK*, usually insufficient
                mul[0] = 4; // 1/4
                }
            else if(w_tot - b_tot <= 1 && w_maj <= 2)
                {
                // no more than 1 minor up, drawish
                mul[0] = 8; // 1/2
                }
            }
        else if(br != 0)
            {
            // assume black sacrifices a rook against the lone pawn
            b_maj--;
            b_tot -= 2;

            if(false) { }
            else if(w_tot == 1)
                {
                // KBK* or KNK*, always insufficient
                mul[0] = 4; // 1/4
                }
            else if(w_tot == 2 && wn == 2)
                {
                // KNNK*, usually insufficient
                mul[0] = 4; // 1/4
                }
            else if(w_tot - b_tot <= 1 && w_maj <= 2)
                {
                // no more than 1 minor up, drawish
                mul[0] = 8; // 1/2
                }
            }
        }

    // black multiplier
    if(bp == 0)
        { // black has no pawns
        int w_maj = wq * 2 + wr;
        int w_min = wb + wn;
        int w_tot = w_maj * 2 + w_min;
        int b_maj = bq * 2 + br;
        int b_min = bb + bn;
        int b_tot = b_maj * 2 + b_min;

        if(false) { }
        else if(b_tot == 1)
            {
            // KBK* or KNK*, always insufficient
            mul[1] = 0;
            }
        else if(b_tot == 2 && bn == 2)
            {
            // KNNK*, usually insufficient
            if(w_tot != 0 || wp == 0)
                {
                mul[1] = 0;
                }
            else
                {           // KNNKP+, might not be draw
                mul[1] = 1; // 1/16
                }
            }
        else if(b_tot == 2 && bb == 2 && w_tot == 1 && wn == 1)
            {
            // KBBKN*, barely drawish (not at all?)
            mul[1] = 8; // 1/2
            }
        else if(b_tot - w_tot <= 1 && b_maj <= 2)
            {
            // no more than 1 minor up, drawish
            mul[1] = 2; // 1/8
            }
        }
    else if(bp == 1)
        { // black has one pawn
        int w_maj = wq * 2 + wr;
        int w_min = wb + wn;
        int w_tot = w_maj * 2 + w_min;
        int b_maj = bq * 2 + br;
        int b_min = bb + bn;
        int b_tot = b_maj * 2 + b_min;

        if(false) { }
        else if(w_min != 0)
            {
            // assume white sacrifices a minor against the lone pawn
            w_min--;
            w_tot--;

            if(false) { }
            else if(b_tot == 1)
                {
                // KBK* or KNK*, always insufficient
                mul[1] = 4; // 1/4
                }
            else if(b_tot == 2 && bn == 2)
                {
                // KNNK*, usually insufficient
                mul[1] = 4; // 1/4
                }
            else if(b_tot - w_tot <= 1 && b_maj <= 2)
                {
                // no more than 1 minor up, drawish
                mul[1] = 8; // 1/2
                }
            }
        else if(wr != 0)
            {
            // assume white sacrifices a rook against the lone pawn
            w_maj--;
            w_tot -= 2;

            if(false) { }
            else if(b_tot == 1)
                {
                // KBK* or KNK*, always insufficient
                mul[1] = 4; // 1/4
                }
            else if(b_tot == 2 && bn == 2)
                {
                // KNNK*, usually insufficient
                mul[1] = 4; // 1/4
                }
            else if(b_tot - w_tot <= 1 && b_maj <= 2)
                {
                // no more than 1 minor up, drawish
                mul[1] = 8; // 1/2
                }
            }
        }

    // potential draw for white
    if(wt == wb + wp && wp >= 1)
        cflags[0] |= (1 << 0);

    if(wt == wb + wp && wb <= 1 && wp >= 1 && bt > bp)
        cflags[0] |= (1 << 1);

    if(wt == 2 && wn == 1 && wp == 1 && bt > bp)
        cflags[0] |= (1 << 2);

    // potential draw for black
    if(bt == bb + bp && bp >= 1)
        cflags[1] |= (1 << 0);

    if(bt == bb + bp && bb <= 1 && bp >= 1 && wt > wp)
        cflags[1] |= (1 << 1);

    if(bt == 2 && bn == 1 && bp == 1 && wt > wp)
        cflags[1] |= (1 << 2);

    // draw leaf (likely draw)
    if(recog == MAT_KQKQ || recog == MAT_KRKR)
        {
        mul[0] = 0;
        mul[1] = 0;
        }

    // king safety
    if(bq >= 1 && bq + br + bb + bn >= 2)
        cflags[0] |= (1 << 3);

    if(wq >= 1 && wq + wr + wb + wn >= 2)
        cflags[1] |= (1 << 3);
    phase = 24;
    phase -= wp * 0;
    phase -= wn * 1;
    phase -= wb * 1;
    phase -= wr * 2;
    phase -= wq * 4;
    phase -= bp * 0;
    phase -= bn * 1;
    phase -= bb * 1;
    phase -= br * 2;
    phase -= bq * 4;

    if(phase < 0)
        phase = 0;
    phase = (phase * 268) / 24;
    // material
    opening = 0;
    endgame = 0;
    owf = wn * 325 + wb * 325 + wr * 500 + wq * 975;
    owf = wn * knight_value_opening + wb * bishop_value_opening + wr * rook_value_opening + wq * queen_value_opening;
    opening += owf;
    opening += wp * pawn_value_opening;
    obf = bn * knight_value_opening + bb * bishop_value_opening + br * rook_value_opening + bq * queen_value_opening;
    opening -= obf;
    opening -= bp * pawn_value_opening;
    ewf = wn * knight_value_endgame + wb * bishop_value_endgame + wr * rook_value_endgame + wq * queen_value_endgame;
    endgame += wp * pawn_value_endgame;
    endgame += ewf;
    ebf = bn * knight_value_endgame + bb * bishop_value_endgame + br * rook_value_endgame + bq * queen_value_endgame;
    endgame -= bp * pawn_value_endgame;
    endgame -= ebf;

    // trade bonus
    if(owf > obf && bp > wp)
        {
        opening += opening_exchange_penalty;
        endgame += opening_exchange_penalty;
        }
    else if(obf > owf && wp > bp)
        {
        opening -= opening_exchange_penalty;
        endgame -= opening_exchange_penalty;
        }

    // bishop pair
    if((wb == 2) && (bb <= 1))
        {
        opening += bishop_pair_value_opening;
        endgame += bishop_pair_value_endgame;

        if(bm == 0)
            {
            opening += 20;
            endgame += 20;
            }
        opening -= (tp * 2);
        endgame -= (tp * 2);
        }

    if((bb == 2) && (wb <= 1))
        {
        opening -= bishop_pair_value_opening;
        endgame -= bishop_pair_value_endgame;

        if(wm == 0)
            {
            opening -= 20;
            endgame -= 20;
            }
        opening += (tp * 2);
        endgame += (tp * 2);
        }

    // material imbalances
    if(use_material_imbalance)
        {
        // rook & bishop vs rook & knight
        if((wr == 1) && (wb == 1) && (wn == 0) && (br == 1) && (bb == 0) && (bn == 1))
            {
            opening += 10 * material_imbalance_weight / 256;
            endgame += 10 * material_imbalance_weight / 256;
            }

        if((br == 1) && (bb == 1) && (bn == 0) && (wr == 1) && (wb == 0) && (wn == 1))
            {
            opening -= 10 * material_imbalance_weight / 256;
            endgame -= 10 * material_imbalance_weight / 256;
            }

        // 2 rooks & 2 bishop vs 2 rooks & 2 knights
        if((wr == 2) && (wb == 2) && (wn == 0) && (br == 2) && (bb == 0) && (bn == 2))
            {
            opening += 10 * material_imbalance_weight / 256;
            endgame += 10 * material_imbalance_weight / 256;
            }

        if((br == 2) && (bb == 2) && (bn == 0) && (wr == 2) && (wb == 0) && (wn == 2))
            {
            opening -= 10 * material_imbalance_weight / 256;
            endgame -= 10 * material_imbalance_weight / 256;
            }

        // bishop or knight vs 3 pawns
        if((wm == bm + 1) && (wp == bp - 3))
            {
            opening += 50 * material_imbalance_weight / 256;
            endgame += 50 * material_imbalance_weight / 256;
            opening += np * 2 * material_imbalance_weight / 256;
            endgame += np * 2 * material_imbalance_weight / 256;
            }

        if((bm == wm + 1) && (bp == wp - 3))
            {
            opening -= 50 * material_imbalance_weight / 256;
            endgame -= 50 * material_imbalance_weight / 256;
            opening -= np * 2 * material_imbalance_weight / 256;
            endgame -= np * 2 * material_imbalance_weight / 256;
            }

        if((wm == bm + 1) && (wp == bp - 4))
            {
            opening += 25 * material_imbalance_weight / 256;
            endgame += 25 * material_imbalance_weight / 256;
            opening += np * 2 * material_imbalance_weight / 256;
            endgame += np * 2 * material_imbalance_weight / 256;
            }

        if((bm == wm + 1) && (bp == wp - 4))
            {
            opening -= 25 * material_imbalance_weight / 256;
            endgame -= 25 * material_imbalance_weight / 256;
            opening -= np * 2 * material_imbalance_weight / 256;
            endgame -= np * 2 * material_imbalance_weight / 256;
            }

        // exchange
        if((wr == br + 1) && (wm == bm - 1))
            {
            if((wm == 3) && (bm == 4))
                {
                opening -= 25 * material_imbalance_weight / 256;
                endgame -= 25 * material_imbalance_weight / 256;
                }

            if((wq == 0) && (wr == 1) && (bq == 0) && (br == 1))
                {
                opening += 25 * material_imbalance_weight / 256;
                endgame += 25 * material_imbalance_weight / 256;
                }
            }

        if((br == wr + 1) && (bm == wm - 1))
            {
            if((bm == 3) && (wm == 4))
                {
                opening += 25 * material_imbalance_weight / 256;
                endgame += 25 * material_imbalance_weight / 256;
                }

            if((bq == 0) && (br == 1) && (wq == 0) && (wr == 1))
                {
                opening -= 25 * material_imbalance_weight / 256;
                endgame -= 25 * material_imbalance_weight / 256;
                }
            }

        // 2 minor pieces vs rook and pawn
        if((wr == br + 1) && (wm == bm - 2) && (wp == bp - 1))
            {
            opening -= np * 2 * material_imbalance_weight / 256;
            endgame -= np * 2 * material_imbalance_weight / 256;

            if(wb == 1)
                {
                opening += 20 * material_imbalance_weight / 256;
                endgame += 20 * material_imbalance_weight / 256;
                }
            }

        if((br == wr + 1) && (bm == wm - 2) && (bp == wp - 1))
            {
            opening += np * 2 * material_imbalance_weight / 256;
            endgame += np * 2 * material_imbalance_weight / 256;

            if(bb == 1)
                {
                opening -= 20 * material_imbalance_weight / 256;
                endgame -= 20 * material_imbalance_weight / 256;
                }
            }

        // rook and knight vs 2 bishops and a pawn
        if((wr == 1) && (wn == 1) && (bb == 2) && (wp == bp - 1))
            {
            if((wq == 0) && (wb == 0) && (bq == 0) && (br == 0) && (bn == 0))
                {
                opening -= 20 * material_imbalance_weight / 256;
                endgame -= 20 * material_imbalance_weight / 256;
                }
            }

        if((br == 1) && (bn == 1) && (wb == 2) && (bp == wp - 1))
            {
            if((bq == 0) && (bb == 0) && (wq == 0) && (wr == 0) && (wn == 0))
                {
                opening += 20 * material_imbalance_weight / 256;
                endgame += 20 * material_imbalance_weight / 256;
                }
            }

        // queen for rook + minor piece + pawn
        if((wq == 1) && (bq == 0) && (wr == br - 1) && (wm == bm - 1) && (wp == bp - 1))
            {
            if(bn > bb)
                {
                opening -= 10 * material_imbalance_weight / 256;
                endgame -= 10 * material_imbalance_weight / 256;
                }
            }

        if((bq == 1) && (wq == 0) && (br == wr - 1) && (bm == wm - 1) && (bp == wp - 1))
            {
            if(wn > wb)
                {
                opening += 10 * material_imbalance_weight / 256;
                endgame += 10 * material_imbalance_weight / 256;
                }
            }

        // queen for 2 rooks
        if((wq == 1) && (wr == 0) && (bq == 0) && (br == 2))
            {
            if(tm == 0)
                {
                opening -= 50 * material_imbalance_weight / 256;
                endgame -= 50 * material_imbalance_weight / 256;
                }

            if(wm == 1 && bm == 1)
                {
                opening -= 25 * material_imbalance_weight / 256;
                endgame -= 25 * material_imbalance_weight / 256;
                }

            if(wm >= 2 && bm >= 2)
                {
                opening += 25 * material_imbalance_weight / 256;
                endgame += 25 * material_imbalance_weight / 256;
                }
            }

        if((bq == 1) && (br == 0) && (wq == 0) && (wr == 2))
            {
            if(tm == 0)
                {
                opening += 50 * material_imbalance_weight / 256;
                endgame += 50 * material_imbalance_weight / 256;
                }

            if(wm == 1 && bm == 1)
                {
                opening += 25 * material_imbalance_weight / 256;
                endgame += 25 * material_imbalance_weight / 256;
                }

            if(wm >= 2 && bm >= 2)
                {
                opening -= 25 * material_imbalance_weight / 256;
                endgame -= 25 * material_imbalance_weight / 256;
                }
            }

        // queen for 3 minor pieces
        if((wq == 1) && (bq == 0) && (wm == bm - 3))
            {
            opening -= 50 * material_imbalance_weight / 256;
            }

        if((bq == 1) && (wq == 0) && (bm == wm - 3))
            {
            opening += 50 * material_imbalance_weight / 256;
            }

        // 2 rooks for 3 minor pieces
        if((wr == 2) && (br == 0) && (wm == bm - 3))
            {
            opening -= 50 * material_imbalance_weight / 256;
            }

        if((br == 2) && (wr == 0) && (bm == wm - 3))
            {
            opening += 50 * material_imbalance_weight / 256;
            }

        // queen and pawns vs rook and 2 minors
        if((wq == 1) && (bq == 0) && (wr == br - 1) && (wm == bm - 2))
            {
            if(wp == bp + 2)
                {
                opening += 25 * material_imbalance_weight / 256;
                endgame += 25 * material_imbalance_weight / 256;
                }

            if(wp == bp + 1)
                {
                opening -= 75 * material_imbalance_weight / 256;
                endgame -= 75 * material_imbalance_weight / 256;
                }
            }

        if((bq == 1) && (wq == 0) && (br == wr - 1) && (bm == wm - 2))
            {
            if(bp == wp + 2)
                {
                opening -= 25 * material_imbalance_weight / 256;
                endgame -= 25 * material_imbalance_weight / 256;
                }

            if(bp == wp + 1)
                {
                opening += 75 * material_imbalance_weight / 256;
                endgame += 75 * material_imbalance_weight / 256;
                }
            }

        // rook vs knight and 2 pawns
        if((wr == br + 1) && (wn == bn - 1) && (wp == bp - 2))
            {
            if(tp < 6)
                {
                opening += 20 * material_imbalance_weight / 256;
                endgame += 20 * material_imbalance_weight / 256;
                }
            else if(tp >= 6)
                {
                opening -= tp * 2 * material_imbalance_weight / 256;
                endgame -= tp * 2 * material_imbalance_weight / 256;
                }
            }

        if((br == wr + 1) && (bn == wn - 1) && (bp == wp - 2))
            {
            if(tp < 6)
                {
                opening -= 20 * material_imbalance_weight / 256;
                endgame -= 20 * material_imbalance_weight / 256;
                }
            else if(tp >= 6)
                {
                opening += tp * 2 * material_imbalance_weight / 256;
                endgame += tp * 2 * material_imbalance_weight / 256;
                }
            }
        }
    // end material imbalances
    // store info
    info->recog = recog;
    info->flags = flags;

    for (color = 0; color < 2; color++)
        info->cflags[color] = cflags[color];

    for (color = 0; color < 2; color++)
        info->mul[color] = mul[color];
    info->phase = phase;
    info->opening = (opening * material_weight) / 256;
    info->endgame = (endgame * material_weight) / 256;
    }
// end of material.cpp
