// search.cpp

// includes
#ifdef _WIN32
#include <windows.h>
#include <process.h>
#else
#include <pthread.h>
#include <errno.h>
#include <unistd.h>
#include <semaphore.h>
#endif

#include <csetjmp>
#include <cstring>
#include "board.h"
#include "book.h"
#include "check.h"
#include "init.h"
#include "option.h"
#include "draw.h"
#include "eval.h"
#include "uci.h"
#include "move.h"
#include "make_move.h"
#include "move_gen.h"
#include "move_list.h"
#include "pawn.h"
#include "probe.h"
#include "pv.h"
#include "search.h"
#include "see.h"
#include "sort.h"
#include "trans.h"
#include "value.h"

#define MIN(a, b)  ((a) < (b) ? (a) : (b))
#define NODE_OPP(type)     (-(type))
#define usable_egbb_value(probe_score,alpha,beta) (((beta > -29744 || probe_score >= 0) \
         && (alpha < 29744 || probe_score <= 0)) || probe_score < -29744)

// variables
static search_multipv_t save_multipv[10];

#ifdef _WIN32
static HANDLE thread_handle[16];
#else
static pthread_t thread_handle[16];
static my_sem_t thread_runnable[16];
#endif

int number_threads;
static int history_value = (history_pruning_threshold * 16384 + 50) / 100;
search_input_t search_input[1];
search_info_t search_info[16][1];
search_root_t search_root[16][1];
search_current_t search_current[16][1];
search_best_t search_best[16][10];

// prototypes
static void search_send_stat(int thread_id);

#ifdef _WIN32
unsigned __stdcall search_thread(void *param);
#else
void *search_thread(void *param);
#endif

static int full_root(list_t *list, board_t *board, int alpha, int beta, int depth, int height, int search_type,
    int thread_id);
static int full_search(board_t *board, int alpha, int beta, int depth, int height, mv_t pv [], int node_type,
    bool extended, int thread_id);
static int full_no_null(board_t *board, int alpha, int beta, int depth, int height, mv_t pv [], int node_type,
    int trans_move, int *best_move, bool extended, int thread_id);
static int full_quiescence(board_t *board, int alpha, int beta, int depth, int height, mv_t pv [], int thread_id);
static int full_new_depth(int depth, int move, board_t *board, bool single_reply, bool in_pv, int height,
    bool extended, bool *cap_extended, int thread_id);
static void pv_fill(const mv_t pv [], board_t *board);
static bool move_is_dangerous(int move, const board_t *board);
static bool capture_is_dangerous(int move, const board_t *board);
static bool simple_stalemate(const board_t *board);
static bool is_passed(const board_t *board, int to);

// functions

// init_threads()
void threads_init()
    {
    int thread_count, option_max;

#ifdef WIN32
    SYSTEM_INFO sys_info;
    GetSystemInfo(&sys_info);
    thread_count = sys_info.dwNumberOfProcessors;
#else
#ifdef _SC_NPROCESSORS_ONLN
    thread_count = sysconf(_SC_NPROCESSORS_ONLN);
#else
    thread_count = 1;
#endif
#endif

    if(thread_count > 16)
        thread_count = 16;
    option_max = option_get_int("Max Threads");

    if(thread_count > option_max)
        thread_count = option_max;
    number_threads = thread_count;
    }

// search_clear()
void search_clear()
    {
    int thread_id;
    // search_input
    search_input->infinite = false;
    search_input->depth_is_limited = false;
    search_input->depth_limit = 0;
    search_input->time_is_limited = false;
    search_input->time_limit_1 = 0.0;
    search_input->time_limit_2 = 0.0;

    // search_info
    for (thread_id = 0; thread_id < number_threads; thread_id++)
        {
        search_info[thread_id]->can_stop = false;
        search_info[thread_id]->stop = false;
        search_info[thread_id]->stopped = false;
        search_info[thread_id]->quiescence_check_nb = 10000;
        search_info[thread_id]->check_inc = 10000;
        search_info[thread_id]->last_time = 0.0;
        // search_best
        search_best[thread_id][search_current[thread_id]->multipv].move = 0;
        search_best[thread_id][search_current[thread_id]->multipv].value = 0;
        search_best[thread_id][search_current[thread_id]->multipv].flags = 0;
        PV_CLEAR(search_best[thread_id][search_current[thread_id]->multipv].pv);
        // search_root
        search_root[thread_id]->depth = 0;
        search_root[thread_id]->move = 0;
        search_root[thread_id]->move_pos = 0;
        search_root[thread_id]->move_nb = 0;
        search_root[thread_id]->last_value = 0;
        search_root[thread_id]->bad_1 = false;
        search_root[thread_id]->bad_2 = false;
        search_root[thread_id]->change = false;
        search_root[thread_id]->easy = false;
        search_root[thread_id]->flag = false;
        // search_current
        search_current[thread_id]->max_depth = 0;
        search_current[thread_id]->node_nb = 0;
        search_current[thread_id]->time = 0.0;
        search_current[thread_id]->speed = 0.0;
        }
    }
#ifdef _WIN32

void start_suspend_threads()
    {
    static int thread_ids[16];
    static unsigned internalthread_ids[16];
    int i;

    // start and suspend threads
    for (i = 1; i < number_threads; i++)
        {
        thread_ids[i - 1] = i;
        thread_handle[i - 1] =
            (HANDLE)_beginthreadex(NULL, 0, &search_thread, &thread_ids[i - 1], CREATE_SUSPENDED,
                &internalthread_ids[i - 1]);
        }
    }

void resume_threads()
    {
    int i;

    // resume threads
    for (i = 1; i < number_threads; i++)
        {
        ResumeThread(thread_handle[i - 1]);
        }
    }
#else

void start_suspend_threads()
    {
    static int thread_ids[16];
    int i;

    // start and suspend threads
    for (i = 1; i < number_threads; i++)
        {
        thread_ids[i - 1] = i;
        my_sem_init(&(thread_runnable[i - 1]), 0);
        pthread_create(&(thread_handle[i - 1]), NULL, search_thread, &(thread_ids[i - 1]));
        }
    }

void resume_threads()
    {
    int i;

    // resume threads
    for (i = 1; i < number_threads; i++)
        {
        my_sem_post(&(thread_runnable[i - 1]));
        }
    }
#endif

// search()

void search()
    {
    int move;
    int i;
    bool all_stopped;
    int thread_id;

    for (i = 0; i < 10; i++)
        {
        save_multipv[i].mate = 0;
        save_multipv[i].depth = 0;
        save_multipv[i].max_depth = 0;
        save_multipv[i].value = 0;
        save_multipv[i].time = 0;
        save_multipv[i].node_nb = 0;
        strcpy(save_multipv[i].pv_string, "");
        }
    search_input->multipv = option_get_int("MultiPV") - 1;

    for (thread_id = 0; thread_id < number_threads; thread_id++)
        {
        search_current[thread_id]->multipv = 0;
        }

    // opening book
    if(option_get_bool("OwnBook") && !search_input->infinite)
        {
        move = book_move(search_input->board);

        if(move != 0)
            {
            // play book move
            search_best[0][search_current[thread_id]->multipv].move = move;
            search_best[0][search_current[thread_id]->multipv].value = 1;
            search_best[0][search_current[thread_id]->multipv].flags = 3;
            search_best[0][search_current[thread_id]->multipv].depth = 1;
            search_best[0][search_current[thread_id]->multipv].pv[0] = move;
            search_best[0][search_current[thread_id]->multipv].pv[1] = 0;
            search_update_best(0);
            return;
            }
        }
    // search_input
    gen_legal_moves(search_input->list, search_input->board);

    if(search_input->search_moves)
        {
        for (int i = 0; i < LIST_SIZE(search_input->list); i++)
            {
            if(!list_contain(search_input->moves_to_search, LIST_MOVE(search_input->list, i)))
                list_remove(search_input->list, i--);
            }
        }

    if(LIST_SIZE(search_input->list) < search_input->multipv + 1)
        {
        search_input->multipv = LIST_SIZE(search_input->list) - 1;
        }

    if(LIST_SIZE(search_input->list) <= 1)
        {
        search_input->depth_is_limited = true;
        search_input->depth_limit = 4;
        }
    trans_inc_date(Trans);
    // resume threads
    resume_threads();
    search_smp(0);

    for (thread_id = 1; thread_id < number_threads; thread_id++)
        { // stop threads
        search_info[thread_id]->stop = true;
        }
    all_stopped = false;

    while(!all_stopped)
        {
        all_stopped = true;

        for (thread_id = 1; thread_id < number_threads; thread_id++)
            {
            if(!search_info[thread_id]->stopped)
                all_stopped = false;
            }
        }
    }
#ifdef _WIN32

unsigned __stdcall search_thread(void *param)
    {
    int thread_id = *((int *)param);

    while(!search_input->exit_engine)
        {
        search_smp(thread_id);
        search_info[thread_id]->stopped = true;
        SuspendThread(thread_handle[thread_id - 1]);
        }
    _endthreadex(0);
    return 0;
    }
#else

void *search_thread(void *param)
    {
    int thread_id = *((int *)param);
    my_sem_wait(&(thread_runnable[thread_id - 1]));

    while(!search_input->exit_engine)
        {
        search_smp(thread_id);
        search_info[thread_id]->stopped = true;
        my_sem_wait(&(thread_runnable[thread_id - 1]));
        }
    return NULL;
    }
#endif

// search_smp()

void search_smp(int thread_id)
    {
    int depth;
    int i;
    bool search_ready;
    sint64 node_nb;
    double speed;

    // search_info
    if(setjmp(search_info[thread_id]->buf) != 0)
        {
        search_update_current(thread_id);
        return;
        }
    // search_root
    list_copy(search_root[thread_id]->list, search_input->list);
    // search_current
    board_copy(search_current[thread_id]->board, search_input->board);
    timer_reset(search_current[thread_id]->timer);
    timer_start(search_current[thread_id]->timer);
    // init
    sort_init(thread_id);
    search_full_init(search_root[thread_id]->list, search_current[thread_id]->board, thread_id);
    // iterative deepening
    search_ready = false;

    if(thread_id == 0)
        { // main thread
        for (depth = 1; depth <= 64; depth++)
            {
            for (search_current[thread_id]->multipv = 0; search_current[thread_id]->multipv <= search_input->multipv;
                search_current[thread_id]->multipv++)
                {
                if(display_depth_start && search_current[thread_id]->multipv == 0)
                    send("info depth %d", depth);
                search_current[thread_id]->act_iteration = depth;
                search_root[thread_id]->bad_1 = false;
                search_root[thread_id]->change = false;
                board_copy(search_current[thread_id]->board, search_input->board);

                if(use_short_search && depth <= short_search_depth)
                    {
                    search_full_root(search_root[thread_id]->list, search_current[thread_id]->board, depth, 1,
                        thread_id);
                    }
                else
                    {
                    search_full_root(search_root[thread_id]->list, search_current[thread_id]->board, depth, 0,
                        thread_id);
                    }
                search_update_current(thread_id);
                node_nb = 0;
                speed = 0;

                for (i = 0; i < number_threads; i++)
                    {
                    node_nb += search_current[thread_id]->node_nb;
                    speed += search_current[thread_id]->speed;
                    }

                if(display_depth_end && search_current[thread_id]->multipv == search_input->multipv)
                    {
                    send("info depth %d time %.0f nodes " S64_FORMAT " nps %.0f",
                        depth, search_current[thread_id]->time*1000.0,node_nb,speed);
                    }

                // update search info
                if(depth >= 1)
                    search_info[thread_id]->can_stop = true;

                if(depth == 1 && LIST_SIZE(search_root[thread_id]->list) >= 2
                    && LIST_VALUE(search_root[thread_id]->list, 0) >= LIST_VALUE(search_root[thread_id]->list, 1)
                    + easy_search_threshold)
                    {
                    search_root[thread_id]->easy = true;
                    }

                if(use_bad && depth > 1)
                    {
                    search_root[thread_id]->bad_2 = search_root[thread_id]->bad_1;
                    search_root[thread_id]->bad_1 = false;
                    }
                search_root[thread_id]->last_value = search_best[thread_id][0].value;

                // stop search?
                if(search_input->depth_is_limited && search_current[thread_id]->multipv >= search_input->multipv
                    && depth >= search_input->depth_limit)
                    {
                    search_root[thread_id]->flag = true;
                    }

                if(search_input->time_is_limited && search_current[thread_id]->time >= search_input->time_limit_1
                    && !search_root[thread_id]->bad_2)
                    {
                    search_root[thread_id]->flag = true;
                    }

                if(use_easy
                && search_input->time_is_limited
                && search_current[thread_id]->time >= search_input->time_limit_1 * easy_search_ratio
                && search_root[thread_id]->easy)
                    {
                    search_root[thread_id]->flag = true;
                    }

                if(use_early_search
                && search_input->time_is_limited
                && search_current[thread_id]->time >= search_input->time_limit_1 * early_search_ratio
                && !search_root[thread_id]->bad_2
                && !search_root[thread_id]->change)
                    {
                    search_root[thread_id]->flag = true;
                    }

                if(search_info[thread_id]->can_stop
                    && (search_info[thread_id]->stop || (search_root[thread_id]->flag && !search_input->infinite)))
                    {
                    search_ready = true;
                    break;
                    }
                }

            if(search_ready)
                break;
            }
        }
    else
        {
        for (depth = 1; depth <= 64; depth++)
            {
            search_current[thread_id]->act_iteration = depth;
            search_info[thread_id]->can_stop = true;
            board_copy(search_current[thread_id]->board, search_input->board);

            if(use_short_search && depth <= short_search_depth)
                {
                search_full_root(search_root[thread_id]->list, search_current[thread_id]->board, depth, 1, thread_id);
                }
            else
                {
                search_full_root(search_root[thread_id]->list, search_current[thread_id]->board, depth, 0, thread_id);
                }
            search_update_current(thread_id);
            }
        }
    }

// search_update_best()
void search_update_best(int thread_id)
    {
    int move, value, flags, depth, max_depth;
    const mv_t *pv;
    double time;
    sint64 node_nb;
    int mate, i, z;
    bool found;
    char move_string[256], pv_string[512];
    search_update_current(thread_id);
#ifdef _WIN32

    EnterCriticalSection(&CS);
#else

    pthread_mutex_lock(&CS);
#endif

    if(display_best && thread_id == 0)
        {
        move = search_best[thread_id][search_current[thread_id]->multipv].move;
        value = search_best[thread_id][search_current[thread_id]->multipv].value;
        flags = search_best[thread_id][search_current[thread_id]->multipv].flags;
        depth = search_best[thread_id][search_current[thread_id]->multipv].depth;
        pv = search_best[thread_id][search_current[thread_id]->multipv].pv;
        max_depth = search_current[thread_id]->max_depth;
        time = search_current[thread_id]->time;
        node_nb = search_current[thread_id]->node_nb;
        move_to_string(move, move_string, 256);
        pv_to_string(pv, pv_string, 512);
        mate = value_to_mate(value);

        if(search_current[thread_id]->multipv == 0)
            {
            save_multipv[search_current[thread_id]->multipv].mate = mate;
            save_multipv[search_current[thread_id]->multipv].depth = depth;
            save_multipv[search_current[thread_id]->multipv].max_depth = max_depth;
            save_multipv[search_current[thread_id]->multipv].value = value;
            save_multipv[search_current[thread_id]->multipv].time = time * 1000.0;
            save_multipv[search_current[thread_id]->multipv].node_nb = node_nb;
            strcpy(save_multipv[search_current[thread_id]->multipv].pv_string, pv_string);
            }
        else
            {
            found = false;

            for (i = 0; i < search_current[thread_id]->multipv; i++)
                {
                if(save_multipv[i].value < value)
                    {
                    found = true;
                    break;
                    }
                }

            if(found)
                {
                for (z = search_current[thread_id]->multipv; z > i; z--)
                    {
                    save_multipv[z].mate = save_multipv[z - 1].mate;
                    save_multipv[z].depth = save_multipv[z - 1].depth;
                    save_multipv[z].max_depth = save_multipv[z - 1].max_depth;
                    save_multipv[z].value = save_multipv[z - 1].value;
                    save_multipv[z].time = save_multipv[z - 1].time;
                    save_multipv[z].node_nb = save_multipv[z - 1].node_nb;
                    strcpy(save_multipv[z].pv_string, save_multipv[z - 1].pv_string);
                    }
                save_multipv[i].mate = mate;
                save_multipv[i].depth = depth;
                save_multipv[i].max_depth = max_depth;
                save_multipv[i].value = value;
                save_multipv[i].time = time * 1000.0;
                save_multipv[i].node_nb = node_nb;
                strcpy(save_multipv[i].pv_string, pv_string);
                }
            else
                {
                save_multipv[search_current[thread_id]->multipv].mate = mate;
                save_multipv[search_current[thread_id]->multipv].depth = depth;
                save_multipv[search_current[thread_id]->multipv].max_depth = max_depth;
                save_multipv[search_current[thread_id]->multipv].value = value;
                save_multipv[search_current[thread_id]->multipv].time = time * 1000.0;
                save_multipv[search_current[thread_id]->multipv].node_nb = node_nb;
                strcpy(save_multipv[search_current[thread_id]->multipv].pv_string, pv_string);
                }
            }

        if(depth > 1 || (depth == 1 && search_current[thread_id]->multipv == search_input->multipv))
            {
            for (i = 0; i <= search_input->multipv; i++)
                {
                if(save_multipv[i].mate == 0)
                    {
                    // normal evaluation
                    if(false) { }
                    else if(flags == 3)
                        {
                        send("info multipv %d depth %d score cp %d time %.0f nodes " S64_FORMAT " pv %s",
                            i + 1, save_multipv[i].depth, save_multipv[i].value, save_multipv[i].time,
                            save_multipv[i].node_nb,save_multipv[i].pv_string);
                        }
                    else if(flags == 2)
                        {
                        send("info multipv %d depth %d score cp %d lowerbound time %.0f nodes " S64_FORMAT " pv %s",
                            i + 1, save_multipv[i].depth, save_multipv[i].value, save_multipv[i].time,
                            save_multipv[i].node_nb,save_multipv[i].pv_string);
                        }
                    else if(flags == 1)
                        {
                        send("info multipv %d depth %d score cp %d upperbound time %.0f nodes " S64_FORMAT " pv %s",
                            i + 1, save_multipv[i].depth, save_multipv[i].value, save_multipv[i].time,
                            save_multipv[i].node_nb,save_multipv[i].pv_string);
                        }
                    }
                else
                    {
                    // mate announcement
                    if(false) { }
                    else if(flags == 3)
                        {
                        send("info multipv %d depth %d score mate %d time %.0f nodes " S64_FORMAT " pv %s",
                            i + 1, save_multipv[i].depth, save_multipv[i].mate, save_multipv[i].time,
                            save_multipv[i].node_nb,save_multipv[i].pv_string);
                        }
                    else if(flags == 2)
                        {
                        send("info multipv %d depth %d score mate %d lowerbound time %.0f nodes " S64_FORMAT " pv %s",
                            i + 1, save_multipv[i].depth,save_multipv[i].mate, save_multipv[i]. time,
                            save_multipv[i].node_nb,save_multipv[i].pv_string);
                        }
                    else if(flags == 1)
                        {
                        send("info multipv %d depth %d score mate %d upperbound time %.0f nodes " S64_FORMAT " pv %s",
                            i + 1, save_multipv[i].depth, save_multipv[i].mate, save_multipv[i].time,
                            save_multipv[i].node_nb,save_multipv[i].pv_string);
                        }
                    }
                }
            }
        }

    if(use_bad && thread_id == 0 && search_best[thread_id][search_current[thread_id]->multipv].depth > 1)
        {
        if(search_best[thread_id][search_current[thread_id]->multipv].value
            <= search_root[thread_id]->last_value - 50)
            {
            search_root[thread_id]->bad_1 = true;
            search_root[thread_id]->easy = false;
            search_root[thread_id]->flag = false;
            }
        else
            {
            search_root[thread_id]->bad_1 = false;
            }
        }
#ifdef _WIN32

    LeaveCriticalSection(&CS);
#else

    pthread_mutex_unlock(&CS);
#endif

    }

// search_update_root()
void search_update_root(int thread_id)
    {
    int move, move_pos, move_nb;
    double time;
    sint64 node_nb;
    char move_string[256];

    if(display_currmove && thread_id == 0)
        {
        search_update_current(thread_id);

        if(search_current[thread_id]->time >= 1.0)
            {
            move = search_root[thread_id]->move;
            move_pos = search_root[thread_id]->move_pos;
            move_nb = search_root[thread_id]->move_nb;
            time = search_current[thread_id]->time;
            node_nb = search_current[thread_id]->node_nb;
            move_to_string(move, move_string, 256);
            send("info currmove %s currmovenumber %d", move_string, move_pos + 1);
            }
        }
    }

// search_update_current()
void search_update_current(int thread_id)
    {
    my_timer_t *timer;
    sint64 node_nb;
    double time, speed;
    timer = search_current[thread_id]->timer;
    node_nb = search_current[thread_id]->node_nb;
    time = (use_cpu_time) ? timer_elapsed_cpu(timer) : timer_elapsed_real(timer);
    speed = (time >= 1.0) ? double(node_nb) / time : 0.0;
    search_current[thread_id]->time = time;
    search_current[thread_id]->speed = speed;
    }

// search_check()
void search_check(int thread_id)
    {
    if(thread_id == 0)
        {
        search_send_stat(thread_id);

        if(use_event)
            event();

        if(search_input->depth_is_limited && search_root[thread_id]->depth > search_input->depth_limit)
            {
            search_root[thread_id]->flag = true;
            }

        if(search_input->time_is_limited && search_current[thread_id]->time >= search_input->time_limit_2)
            {
            search_root[thread_id]->flag = true;
            }

        if(search_input->time_is_limited && search_current[thread_id]->time >= search_input->time_limit_1
            && !search_root[thread_id]->bad_1 && !search_root[thread_id]->bad_2
            && (!use_extension || search_root[thread_id]->move_pos == 0))
            {
            search_root[thread_id]->flag = true;
            }

        if(search_info[thread_id]->can_stop
            && (search_info[thread_id]->stop || (search_root[thread_id]->flag && !search_input->infinite)))
            {
            longjmp(search_info[thread_id]->buf, 1);
            }
        }
    else
        {
        if(search_info[thread_id]->stop)
            {
            longjmp(search_info[thread_id]->buf, 1);
            }
        }
    }

// search_send_stat()
static void search_send_stat(int thread_id)
    {
    double time, speed;
    sint64 node_nb;
    int i;
    search_update_current(thread_id);

    if(display_stats && thread_id == 0 && search_current[thread_id]->time >= search_info[thread_id]->last_time + 1.0)
        { // at least one-second gap
        search_info[thread_id]->last_time = search_current[thread_id]->time;
        time = search_current[thread_id]->time;
        speed = search_current[thread_id]->speed;
        node_nb = 0;
        speed = 0;

        for (i = 0; i < number_threads; i++)
            {
            node_nb += search_current[thread_id]->node_nb;
            speed += search_current[thread_id]->speed;
            }
        send("info time %.0f nodes " S64_FORMAT " nps %.0f",time*1000.0,node_nb,speed);
        trans_stats(Trans);
        }
    }

// search_full_init()
void search_full_init(list_t *list, board_t *board, int thread_id)
    {
    int trans_move, trans_table_depth, trans_flags, trans_value;
    entry_t *found_entry;
    // standard sort
    list_note(list);
    list_sort(list);
    // basic sort
    trans_move = 0;

    if(use_trans_table)
        trans_retrieve(Trans, &found_entry, board->key, &trans_move, &trans_table_depth, &trans_flags, &trans_value);
    note_moves(list, board, 0, trans_move, thread_id);
    list_sort(list);
    }

// search_full_root()
int search_full_root(list_t *list, board_t *board, int depth, int search_type, int thread_id)
    {
    int value, a, b;

    if(depth < 3)
        {
        a = -30000;
        b = +30000;
        }
    else
        {
        if(search_root[thread_id]->last_value == 0)
            {
            a = search_root[thread_id]->last_value - 1;
            b = search_root[thread_id]->last_value + 1;
            }
        else
            {
            a = search_root[thread_id]->last_value - window_size;
            b = search_root[thread_id]->last_value + window_size;
            }
        }
    value = full_root(list, board, a, b, depth, 0, search_type, thread_id);
    return value;
    }

// full_root()
static int full_root(list_t *list, board_t *board, int alpha, int beta, int depth, int height, int search_type,
    int thread_id)
    {
    int old_alpha;
    int value, best_value[10];
    int i, move, j;
    int new_depth;
    undo_t undo[1];
    mv_t new_pv[256];
    bool found;
    bool cap_extended;
    bool reduced;
    // init
    search_current[thread_id]->node_nb++;
    search_info[thread_id]->quiescence_check_nb--;

    if(search_current[thread_id]->multipv == 0)
        for (i = 0; i < LIST_SIZE(list); i++)
            list->value[i] = -32767;
    old_alpha = alpha;
    best_value[search_current[thread_id]->multipv] = -32767;

    // move loop
    for (i = 0; i < LIST_SIZE(list); i++)
        {
        move = LIST_MOVE(list, i);

        if(search_current[thread_id]->multipv > 0)
            {
            found = false;

            for (j = 0; j < search_current[thread_id]->multipv; j++)
                {
                if(search_best[thread_id][j].pv[0] == move)
                    {
                    found = true;
                    break;
                    }
                }

            if(found == true)
                continue;
            }
        search_root[thread_id]->depth = depth;
        search_root[thread_id]->move = move;
        search_root[thread_id]->move_pos = i;
        search_root[thread_id]->move_nb = LIST_SIZE(list);
        search_update_root(thread_id);
        new_depth =
            full_new_depth(depth, move, board, board_is_check(board) && LIST_SIZE(list) == 1, true, height, false,
                &cap_extended, thread_id);
        reduced = false;

        if(i > 10 && search_type == 0 && depth >= 3 && new_depth < depth && !move_is_tactical(move, board))
            {
            new_depth--;
            reduced = true;
            }
        move_do(board, move, undo);

        if(search_type == 1 || best_value[search_current[thread_id]->multipv] == -32767)
            { // first move
            value = -full_search(board, -beta, -alpha, new_depth, height + 1, new_pv, 0, cap_extended, thread_id);

            if(value <= alpha)
                { // research
                value = -full_search(board, -alpha, 30000, new_depth, height + 1, new_pv, 0, cap_extended, thread_id);
                old_alpha = alpha = -30000;
                }
            else if(value >= beta)
                {
                value = -full_search(board, -30000, -beta, new_depth, height + 1, new_pv, 0, cap_extended, thread_id);
                beta = +30000;
                }
            }
        else
            { // other moves
            value = -full_search(board, -alpha - 1, -alpha, new_depth, height + 1, new_pv, 1, cap_extended, thread_id);

            if(value > alpha)
                { // && value < beta
                search_root[thread_id]->change = true;
                search_root[thread_id]->easy = false;
                search_root[thread_id]->flag = false;

                if(reduced)
                    new_depth++;
                value = -full_search(board, -30000, -alpha, new_depth, height + 1, new_pv, 0, cap_extended, thread_id);

                if(value >= beta)
                    {
                    beta = +30000;
                    }
                }
            }
        move_undo(board, move, undo);
        list->value[i] = value;

        if(value > best_value[search_current[thread_id]->multipv]
            && (best_value[search_current[thread_id]->multipv] == -32767 || value > alpha))
            {
            search_best[thread_id][search_current[thread_id]->multipv].move = move;
            search_best[thread_id][search_current[thread_id]->multipv].value = value;

            if(value <= alpha)
                { // upper bound
                search_best[thread_id][search_current[thread_id]->multipv].flags = 1;
                }
            else if(value >= beta)
                { // lower bound
                search_best[thread_id][search_current[thread_id]->multipv].flags = 2;
                }
            else
                { // alpha < value < beta => exact value
                search_best[thread_id][search_current[thread_id]->multipv].flags = 3;
                }
            search_best[thread_id][search_current[thread_id]->multipv].depth = depth;
            pv_cat(search_best[thread_id][search_current[thread_id]->multipv].pv, new_pv, move);
            search_update_best(thread_id);
            }

        if(value > best_value[search_current[thread_id]->multipv])
            {
            best_value[search_current[thread_id]->multipv] = value;

            if(value > alpha)
                {
                if(search_type == 0)
                    alpha = value;
                }
            }
        }
    list_sort(list);

    if(use_trans_table && best_value[search_current[thread_id]->multipv] > old_alpha
        && best_value[search_current[thread_id]->multipv] < beta)
        {
        pv_fill(search_best[thread_id][search_current[thread_id]->multipv].pv, board);
        }
    good_move(search_best[thread_id]->move, board, depth, height, thread_id);
    return best_value[search_current[thread_id]->multipv];
    }

// full_search()
static int full_search(board_t *board, int alpha, int beta, int depth, int height, mv_t pv [], int node_type,
    bool extended, int thread_id)
    {
    bool in_check;
    bool single_reply;
    bool good_cap;
    int trans_move, trans_table_depth, trans_flags, trans_value;
    int old_alpha;
    int value, best_value;
    int move, best_move;
    int new_depth;
    int played_nb;
    int i;
    int opt_value;
    bool reduced, cap_extended;
    int probe_score;
    int futility_pruning_margin;
    int probe_depth;
    attack_t attack[1];
    sort_t sort[1];
    undo_t undo[1];
    mv_t new_pv[256];
    mv_t played[256];
    entry_t *found_entry;

    // horizon?
    if(depth <= 0)
        {
        return full_quiescence(board, alpha, beta, 0, height, pv, thread_id);
        }
    // init
    search_current[thread_id]->node_nb++;
    search_info[thread_id]->quiescence_check_nb--;
    PV_CLEAR(pv);

    if(height > search_current[thread_id]->max_depth)
        search_current[thread_id]->max_depth = height;

    if(search_info[thread_id]->quiescence_check_nb <= 0)
        {
        search_info[thread_id]->quiescence_check_nb += search_info[thread_id]->check_inc;
        search_check(thread_id);
        }

    // draw?
    if(board_is_repetition(board))
        return 0;

    if(egbb_is_loaded && board->piece_nb <= 5)
        {
        probe_depth = (2 * search_current[thread_id]->act_iteration) / 3;

        if((board->piece_nb <= 4 || height <= probe_depth)
            && (height >= probe_depth || board->cap_sq != 0 || PIECE_IS_PAWN(board->moving_piece)))
            {
            if(probe_bitbases(board, probe_score))
                {
                if(probe_score != 0 && board_is_mate(board))
                    return (probe_score < 0 ? 1 : -1) * VALUE_MATE(height);
                else
                    {
                    probe_score = value_from_trans(probe_score, height);
                    return probe_score;
                    }
                }
            }
        }

    if(recog_draw(board, thread_id))
        return 0;

    if(use_distance_pruning)
        {
        // mate-distance pruning
        // lower bound
        value = VALUE_MATE(height + 2); // does not work if the current position is mate

        if(value > alpha && board_is_mate(board))
            value = VALUE_MATE(height);

        if(value > alpha)
            {
            alpha = value;

            if(value >= beta)
                return value;
            }

        // upper bound
        value = -VALUE_MATE(height + 1);

        if(value < beta)
            {
            beta = value;

            if(value <= alpha)
                return value;
            }
        }

    // transposition table
    trans_move = 0;

    if(use_trans_table && depth >= 1)
        {
        if(trans_retrieve(Trans, &found_entry, board->key, &trans_move, &trans_table_depth, &trans_flags,
            &trans_value))
            {
            if(node_type != 0)
                {
                if(use_mate_values)
                    {
                    if(trans_table_depth < depth)
                        {
                        if(trans_value < -(29744) && TRANS_IS_UPPER(trans_flags))
                            {
                            trans_table_depth = depth;
                            trans_flags = (1 << 0);
                            }
                        else if(trans_value > +(29744) && TRANS_IS_LOWER(trans_flags))
                            {
                            trans_table_depth = depth;
                            trans_flags = (1 << 1);
                            }
                        }
                    }

                if(trans_flags == 1 << 2 && board->piece_nb == 5 && usable_egbb_value(trans_value, alpha, beta))
                    return value_from_trans(trans_value, height);

                if(trans_table_depth >= depth)
                    {
                    trans_value = value_from_trans(trans_value, height);

                    if((use_exact && TRANS_IS_EXACT(trans_flags))
                        || (TRANS_IS_LOWER(trans_flags) && trans_value >= beta)
                            || (TRANS_IS_UPPER(trans_flags) && trans_value <= alpha))
                        {
                        return trans_value;
                        }
                    }
                }
            }
        }

    // height limit
    if(height >= 255)
        return eval(board, alpha, beta, thread_id);
    // more init
    old_alpha = alpha;
    best_value = -32767;
    best_move = 0;
    played_nb = 0;
    attack_set(attack, board);
    in_check = ATTACK_IN_CHECK(attack);

    // null-move pruning
    if(use_null_move && depth >= null_move_depth && node_type != 0)
        {
        if(!in_check && !value_is_mate(beta) && (board->piece_size[board->turn] >= 2)
            && (!use_null_move_eval || depth <= null_move_reduction
                + 1 || eval(board, alpha, beta, thread_id) >= beta))
            {
            if((depth > 7) && (board->piece_size[board->turn] >= 3))
                null_move_reduction = 4;
            else
                null_move_reduction = 3;
            new_depth = depth - null_move_reduction - 1;
            move_do_null(board, undo);
            value =
                -full_search(board, -beta, -beta + 1, new_depth, height + 1, new_pv, NODE_OPP(node_type), false,
                    thread_id);
            move_undo_null(board, undo);

            // verification search
            if(use_verification_search && depth > verification_search_reduction)
                {
                if(value >= beta && (!use_verification_search_endgame || board->piece_size[board->turn] <= 3))
                    {
                    new_depth = depth - verification_search_reduction;
                    int ver_value =
                        full_no_null(board, alpha, beta, new_depth, height, new_pv, 1, trans_move, &move, false,
                            thread_id);

                    if(value >= beta)
                        {
                        played[played_nb++] = move;
                        best_move = move;
                        best_value = MIN(value, ver_value);
                        pv_copy(pv, new_pv);
                        goto cut;
                        }
                    }
                }

            // pruning
            if(value >= beta)
                {
                if(value > +(29744))
                    value = +(29744); // do not return unproven mates
                best_move = 0;
                best_value = value;
                goto cut;
                }
            }
        }
    // razoring
    else if(use_razoring && node_type != 0 && depth <= razor_depth
        && eval(board, alpha, beta, thread_id) < beta - razor_margin)
        {
        value = full_quiescence(board, alpha, beta, 0, height, pv, thread_id);

        if(value < beta)
            return value;
        }

    // internal iterative deepening
    if(use_iid && depth >= iid_depth && node_type == 0 && trans_move == 0)
        {
        new_depth = MIN(depth - iid_reduction, depth / 2);
        value = full_search(board, alpha, beta, new_depth, height, new_pv, node_type, false, thread_id);

        if(value <= alpha)
            value = full_search(board, -30000, beta, new_depth, height, new_pv, node_type, false, thread_id);
        trans_move = new_pv[0];
        }
    // move generation
    sort_init(sort, board, attack, depth, height, trans_move, thread_id);
    single_reply = false;

    if(in_check && LIST_SIZE(sort->list) == 1)
        single_reply = true;
    // move loop
    opt_value = +30000;
    good_cap = true;

    while((move = sort_next(sort, thread_id)) != 0)
        {
        // extensions
        new_depth =
            full_new_depth(depth, move, board, single_reply, node_type == 0, height, extended, &cap_extended,
                thread_id);
        // history pruning
        bool can_reduce = !in_check && new_depth < depth && !move_is_dangerous(move, board);
        value = sort->value; // history score

        if(use_history_pruning && can_reduce && depth <= 4 && node_type != 0 && value < history_value / (depth / 3 + 1)
            && played_nb >= 1
            + depth)
            {
            continue;
            }

        // futility pruning
        if(use_futility_pruning && can_reduce && node_type != 0 && depth <= 3)
            {
            if(!in_check && new_depth < depth && !move_is_tactical(move, board) && !move_is_check(move, board)
                && !move_is_dangerous(move, board))
                {
                if(opt_value == +30000)
                    {
                    if(depth >= 2)
                        futility_pruning_margin = 300 + (depth - 2) * 50;
                    else
                        futility_pruning_margin = 50;
                    opt_value = eval(board, 32768, -32768, thread_id) + futility_pruning_margin;
                    }
                value = opt_value;

                // pruning
                if(value <= alpha)
                    {
                    if(value > best_value)
                        {
                        best_value = value;
                        PV_CLEAR(pv);
                        }
                    continue;
                    }
                }
            }
        reduced = false;

        if(good_cap && !move_is_tactical(move, board))
            {
            good_cap = false;
            }

        if(use_history_pruning && can_reduce && played_nb >= history_pruning_move_nb
            && depth >= history_pruning_depth)
            {
            if(!good_cap)
                {
                if(node_type == -1)
                    {
                    new_depth--;
                    reduced = true;
                    }
                else
                    {
                    if(played_nb >= history_pruning_move_nb + 3)
                        {
                        new_depth--;
                        reduced = true;
                        }
                    }
                }
            }
        // recursive search
        move_do(board, move, undo);

        if(node_type != 0 || best_value == -32767)
            { // first move
            value =
                -full_search(board, -beta, -alpha, new_depth, height + 1, new_pv, NODE_OPP(node_type), cap_extended,
                    thread_id);

            if(value > alpha && reduced)
                {
                new_depth++;
                value =
                    -full_search(board, -beta, -alpha, new_depth, height + 1, new_pv, NODE_OPP(node_type), cap_extended,
                        thread_id);
                }
            }
        else
            { // other moves
            value = -full_search(board, -alpha - 1, -alpha, new_depth, height + 1, new_pv, 1, cap_extended, thread_id);

            if(value > alpha)
                {
                if(reduced)
                    new_depth++;
                value = -full_search(board, -beta, -alpha, new_depth, height + 1, new_pv, 0, cap_extended, thread_id);
                }
            }
        move_undo(board, move, undo);
        played[played_nb++] = move;

        if(value > best_value)
            {
            best_value = value;
            pv_cat(pv, new_pv, move);

            if(value > alpha)
                {
                alpha = value;
                best_move = move;

                if(value >= beta)
                    {
                    goto cut;
                    }
                }
            }
        }

    // ALL node
    if(best_value == -32767)
        { // no legal move
        if(in_check)
            {
            return VALUE_MATE(height);
            }
        else
            {
            return 0;
            }
        }
    cut:
    // move ordering
    if(best_move != 0)
        {
        good_move(best_move, board, depth, height, thread_id);

        if(best_value >= beta && !move_is_tactical(best_move, board))
            {
            for (i = 0; i < played_nb - 1; i++)
                {
                move = played[i];
                history_bad(move, board, thread_id);
                }
            history_good(best_move, board, thread_id);
            }
        }

    // transposition table
    if(use_trans_table && depth >= trans_table_depth)
        {
        trans_move = best_move;
        trans_table_depth = depth;
        trans_flags = 0;

        if(best_value > old_alpha)
            trans_flags |= (1 << 1);

        if(best_value < beta)
            trans_flags |= (1 << 0);
        trans_value = value_to_trans(best_value, height);
        trans_store(Trans, board->key, trans_move, trans_table_depth, trans_flags, trans_value);
        }
    return best_value;
    }

// full_no_null()
static int full_no_null(board_t *board, int alpha, int beta, int depth, int height, mv_t pv [], int node_type,
    int trans_move, int *best_move, bool extended, int thread_id)
    {
    int value, best_value;
    int move;
    int new_depth;
    attack_t attack[1];
    sort_t sort[1];
    undo_t undo[1];
    mv_t new_pv[256];
    bool cap_extended;
    // init
    search_current[thread_id]->node_nb++;
    search_info[thread_id]->quiescence_check_nb--;
    PV_CLEAR(pv);

    if(height > search_current[thread_id]->max_depth)
        search_current[thread_id]->max_depth = height;

    if(search_info[thread_id]->quiescence_check_nb <= 0)
        {
        search_info[thread_id]->quiescence_check_nb += search_info[thread_id]->check_inc;
        search_check(thread_id);
        }
    attack_set(attack, board);
    *best_move = 0;
    best_value = -32767;
    // move loop
    sort_init(sort, board, attack, depth, height, trans_move, thread_id);

    while((move = sort_next(sort, thread_id)) != 0)
        {
        new_depth = full_new_depth(depth, move, board, false, false, height, extended, &cap_extended, thread_id);
        move_do(board, move, undo);
        value =
            -full_search(board, -beta, -alpha, new_depth, height + 1, new_pv, NODE_OPP(node_type), cap_extended,
                thread_id);
        move_undo(board, move, undo);

        if(value > best_value)
            {
            best_value = value;
            pv_cat(pv, new_pv, move);

            if(value > alpha)
                {
                alpha = value;
                *best_move = move;

                if(value >= beta)
                    goto cut;
                }
            }
        }

    // ALL node
    if(best_value == -32767)
        { // no legal move => stalemate
        best_value = 0;
        }
    cut:
    return best_value;
    }

// full_quiescence()
static int full_quiescence(board_t *board, int alpha, int beta, int depth, int height, mv_t pv [], int thread_id)
    {
    bool in_check;
    int old_alpha;
    int value, best_value;
    int best_move;
    int move;
    int opt_value;
    attack_t attack[1];
    sort_t sort[1];
    undo_t undo[1];
    mv_t new_pv[256];
    // init
    search_current[thread_id]->node_nb++;
    search_info[thread_id]->quiescence_check_nb--;
    PV_CLEAR(pv);

    if(height > search_current[thread_id]->max_depth)
        search_current[thread_id]->max_depth = height;

    if(search_info[thread_id]->quiescence_check_nb <= 0)
        {
        search_info[thread_id]->quiescence_check_nb += search_info[thread_id]->check_inc;
        search_check(thread_id);
        }

    // draw?
    if(board_is_repetition(board))
        return 0;

    if(recog_draw(board, thread_id))
        return 0;

    // mate-distance pruning
    if(use_distance_pruning)
        {
        // lower bound
        value = VALUE_MATE(height + 2);

        if(value > alpha && board_is_mate(board))
            value = VALUE_MATE(height);

        if(value > alpha)
            {
            alpha = value;

            if(value >= beta)
                return value;
            }

        // upper bound
        value = -VALUE_MATE(height + 1);

        if(value < beta)
            {
            beta = value;

            if(value <= alpha)
                return value;
            }
        }

    // more init
    attack_set(attack, board);
    in_check = ATTACK_IN_CHECK(attack);

    if(in_check)
        {
        depth++;
        }

    // height limit
    if(height >= 255)
        return eval(board, alpha, beta, thread_id);
    // more init
    old_alpha = alpha;
    best_value = -32767;
    best_move = 0;
    opt_value = +30000;

    if(!in_check)
        {
        // lone-king stalemate?
        if(simple_stalemate(board))
            return 0;
        // stand pat
        value = eval(board, alpha, beta, thread_id);
        best_value = value;

        if(value > alpha)
            {
            alpha = value;

            if(value >= beta)
                goto cut;
            }

        if(use_delta_pruning)
            {
            opt_value = value + 50;
            }
        }
    sort_init_qs(sort, board, attack, depth >= -(alpha != beta - 1));

    while((move = sort_next_qs(sort)) != 0)
        {
        // delta pruning
        if(use_delta_pruning && beta == old_alpha + 1)
            {
            if(!in_check && !move_is_check(move, board) && !capture_is_dangerous(move, board))
                {
                // optimistic evaluation
                value = opt_value;
                int to = MOVE_TO(move);
                int capture = board->square[to];

                if(capture != 0)
                    {
                    value += VALUE_PIECE(capture);
                    }
                else if(MOVE_IS_EN_PASSANT(move))
                    {
                    value += value_pawn;
                    }

                if(MOVE_IS_PROMOTE(move))
                    value += value_queen - value_pawn;

                if(PIECE_IS_BISHOP(capture))
                    {
                    if(board->turn == 0 && board->number[5] >= 2)
                        value += 50;

                    if(board->turn == 1 && board->number[4] >= 2)
                        value += 50;
                    }

                // pruning
                if(value <= alpha)
                    {
                    if(value > best_value)
                        {
                        best_value = value;
                        PV_CLEAR(pv);
                        }
                    continue;
                    }
                }
            }
        move_do(board, move, undo);
        value = -full_quiescence(board, -beta, -alpha, depth - 1, height + 1, new_pv, thread_id);
        move_undo(board, move, undo);

        if(value > best_value)
            {
            best_value = value;
            pv_cat(pv, new_pv, move);

            if(value > alpha)
                {
                alpha = value;
                best_move = move;

                if(value >= beta)
                    goto cut;
                }
            }
        }

    // ALL node
    if(best_value == -32767)
        { // no legal move
        return VALUE_MATE(height);
        }
    cut:
    return best_value;
    }

// full_new_depth()
static int full_new_depth(int depth, int move, board_t *board, bool single_reply, bool in_pv, int height,
    bool extended, bool *cap_extended, int thread_id)
    {
    *cap_extended = false;
    int to = MOVE_TO(move);
    int capture = board->square[to];
    int piece_nb = board->piece_size[0] + board->piece_size[1];

    if(in_pv && piece_nb <= 5 && capture != 0 && !PIECE_IS_PAWN(capture))
        return depth;

    if(single_reply)
        return depth;

    if(move_is_check(move, board) && (in_pv || see_move(move, board) >= -100))
        return depth;

    if(in_pv && move_is_dangerous(move, board))
        return depth;

    if(in_pv && capture != 0 && !extended && see_move(move, board) >= -100)
        {
        *cap_extended = true;
        return depth;
        }
    return depth - 1;
    }

// pv_fill()
static void pv_fill(const mv_t pv [], board_t *board)
    {
    int move;
    int trans_move, trans_table_depth;
    undo_t undo[1];
    move = *pv;

    if(move != 0 && move != 11)
        {
        move_do(board, move, undo);
        pv_fill(pv + 1, board);
        move_undo(board, move, undo);
        trans_move = move;
        trans_table_depth = -127;
        trans_store(Trans, board->key, trans_move, trans_table_depth, 0, -30000);
        }
    }

// move_is_dangerous()
static bool move_is_dangerous(int move, const board_t *board)
    {
    int piece;
    piece = MOVE_PIECE(move, board);

    if(PIECE_IS_PAWN(piece) && is_passed(board, MOVE_TO(move)))
        {
        return true;
        }
    return false;
    }

// capture_is_dangerous()
static bool capture_is_dangerous(int move, const board_t *board)
    {
    int piece, capture;
    piece = MOVE_PIECE(move, board);

    if(PIECE_IS_PAWN(piece) && PAWN_RANK(MOVE_TO(move), board->turn) >= (0xA))
        {
        return true;
        }
    capture = move_capture(move, board);

    if(PIECE_IS_QUEEN(capture))
        return true;

    if(PIECE_IS_PAWN(capture) && PAWN_RANK(MOVE_TO(move), board->turn) <= (0x5))
        {
        return true;
        }
    return false;
    }

// simple_stalemate()
static bool simple_stalemate(const board_t *board)
    {
    int me, opp;
    int king;
    int opp_flag;
    int from, to;
    int capture;
    const inc_t *inc_ptr;
    int inc;
    // lone king?
    me = board->turn;

    if(board->piece_size[me] != 1 || board->pawn_size[me] != 0)
        return false; // no
    // king in a corner?
    king = KING_POS(board, me);

    if(king != (0x44) && king != (0x4B) && king != (0xB4) && king != (0xBB))
        return false; // no
    // init
    opp = COLOR_OPP(me);
    opp_flag = COLOR_FLAG(opp);
    // king can move?
    from = king;

    for (inc_ptr = king_inc; (inc = *inc_ptr) != 0; inc_ptr++)
        {
        to = from + inc;
        capture = board->square[to];

        if(capture == 0 || FLAG_IS(capture, opp_flag))
            {
            if(!is_attacked(board, to, opp))
                return false; // legal king move
            }
        }
    // no legal move
    return true;
    }

static bool is_passed(const board_t *board, int to)
    {
    int t2;
    int me, opp;
    int file, rank;
    me = board->turn;
    opp = COLOR_OPP(me);
    file = SQUARE_FILE(to);
    rank = PAWN_RANK(to, me);
    t2 = board->pawn_file[me][file] | bit_rev[board->pawn_file[opp][file]];

    // passed pawns
    if((t2 &bit_gt[rank]) == 0)
        {
        if(((bit_rev[board->pawn_file[opp][file - 1]] | bit_rev[board->pawn_file[opp][file + 1]]) & bit_gt[rank])
            == 0)
            {
            return true;
            }
        }
    return false;
    }
// end of search.cpp
