// sort.cpp

// includes
#include "move_list.h"
#include "check.h"
#include "evasion.h"
#include "move_gen.h"
#include "valid_move.h"
#include "search.h"
#include "see.h"
#include "sort.h"
#include "value.h"

// macros
#define MIN(a,b) (a < b ? a : b)
#define HISTORY_INC(depth) ((depth)*(depth))

// types
enum gen_t
    {
    GEN_ERROR,
    GEN_LEGAL_EVASION,
    GEN_TRANS,
    GEN_GOOD_CAPTURE,
    GEN_BAD_CAPTURE,
    GEN_KILLER,
    GEN_QUIET,
    GEN_EVASION_QS,
    GEN_CAPTURE_QS,
    GEN_CHECK_QS,
    GEN_END
    };

enum test_t
    {
    TEST_ERROR,
    TEST_NONE,
    TEST_LEGAL,
    TEST_TRANS_KILLER,
    TEST_GOOD_CAPTURE,
    TEST_BAD_CAPTURE,
    TEST_KILLER,
    TEST_QUIET,
    TEST_CAPTURE_QS,
    TEST_CHECK_QS
    };

// variables
static int pos_legal_evasion;
static int pos_see;
static int pos_evasionqs;
static int pos_check_qs;
static int pos_capture_qs;
static int code[256];
static uint16 killer[16][256][2];
static uint16 history[16][768];
static uint16 hist_hit[16][768];
static uint16 hist_tot[16][768];

// prototypes
static void note_quiet_moves(list_t *list, const board_t *board, int thread_id);
static void note_moves_simple(list_t *list, const board_t *board);
static void note_mvv_lva(list_t *list, const board_t *board);
static int move_value(int move, const board_t *board, int height, int trans_killer, int thread_id);
static int capture_value(int move, const board_t *board);
static int quiet_move_value(int move, const board_t *board, int thread_id);
static int move_value_simple(int move, const board_t *board);
static int history_prob(int move, const board_t *board, int thread_id);
static void store_killer(int move, int height, int thread_id);
static int killer_number(int move, int height, int thread_id);
static bool move_is_killer(int move, int killer []);
static bool capture_is_good(int move, const board_t *board);
static int mvv_lva(int move, const board_t *board);
static uint16 history_index(int move, const board_t *board);

// functions

// sort_init()
void sort_init(int thread_id)
    {
    int i, height;
    int pos;

    // killer
    for (height = 0; height < 256; height++)
        {
        for (i = 0; i < 2; i++)
            killer[thread_id][height][i] = 0;
        }

    // history
    for (i = 0; i < 768; i++)
        history[thread_id][i] = 0;

    for (i = 0; i < 768; i++)
        {
        hist_hit[thread_id][i] = 1;
        hist_tot[thread_id][i] = 1;
        }

    // code[]
    for (pos = 0; pos < 256; pos++)
        code[pos] = GEN_ERROR;
    pos = 0;
    // main search
    pos_legal_evasion = pos;
    code[pos++] = GEN_LEGAL_EVASION;
    code[pos++] = GEN_END;
    pos_see = pos;
    code[pos++] = GEN_TRANS;
    code[pos++] = GEN_GOOD_CAPTURE;
    code[pos++] = GEN_KILLER;
    code[pos++] = GEN_QUIET;
    code[pos++] = GEN_BAD_CAPTURE;
    code[pos++] = GEN_END;
    // quiescence search
    pos_evasionqs = pos;
    code[pos++] = GEN_EVASION_QS;
    code[pos++] = GEN_END;
    pos_check_qs = pos;
    code[pos++] = GEN_CAPTURE_QS;
    code[pos++] = GEN_CHECK_QS;
    code[pos++] = GEN_END;
    pos_capture_qs = pos;
    code[pos++] = GEN_CAPTURE_QS;
    code[pos++] = GEN_END;
    }

// sort_init()
void sort_init(sort_t *sort, board_t *board, const attack_t *attack, int depth, int height, int trans_killer,
    int thread_id)
    {
    int i;
    sort->board = board;
    sort->attack = attack;
    sort->depth = depth;
    sort->height = height;
    sort->capture_nb = 0;
    sort->trans_killer = trans_killer;

    for (i = 0; i < 2; i++)
        sort->killer[i] = killer[thread_id][height][i];

    if(ATTACK_IN_CHECK(sort->attack))
        {
        gen_legal_evasions(sort->list, sort->board, sort->attack);
        note_moves(sort->list, sort->board, sort->height, sort->trans_killer, thread_id);
        list_sort(sort->list);
        sort->gen = pos_legal_evasion + 1;
        sort->test = TEST_NONE;
        }
    else
        { // not in check
        LIST_CLEAR(sort->list);
        sort->gen = pos_see;
        }
    sort->pos = 0;
    }

// sort_next()
int sort_next(sort_t *sort, int thread_id)
    {
    int move;
    int gen;
    int i;

    while(true)
        {
        while(sort->pos < LIST_SIZE(sort->list))
            {
            // next move
            move = LIST_MOVE(sort->list, sort->pos);
            sort->value = 16384; // default score
            sort->valuePV = 16384;
            sort->pos++;

            // test
            if(false) { }
            else if(sort->test == TEST_NONE) {
            // no-op
            }
            else if(sort->test == TEST_TRANS_KILLER)
                {
                if(!move_is_pseudo(move, sort->board))
                    continue;

                if(!pseudo_is_legal(move, sort->board))
                    continue;
                }
            else if(sort->test == TEST_GOOD_CAPTURE)
                {
                if(move == sort->trans_killer)
                    continue;

                if(!capture_is_good(move, sort->board))
                    {
                    LIST_ADD(sort->bad, move);
                    continue;
                    }

                if(!pseudo_is_legal(move, sort->board))
                    continue;
                }
            else if(sort->test == TEST_BAD_CAPTURE)
                {
                if(!pseudo_is_legal(move, sort->board))
                    continue;
                }
            else if(sort->test == TEST_KILLER)
                {
                if(move == sort->trans_killer)
                    continue;

                if(!quiet_is_pseudo(move, sort->board))
                    continue;

                if(!pseudo_is_legal(move, sort->board))
                    continue;

                if(killer_number(move, sort->height, thread_id) >= killer_no_pruning)
                    sort->value = history_prob(move, sort->board, thread_id);
                }
            else if(sort->test == TEST_QUIET)
                {
                if(move == sort->trans_killer)
                    continue;

                if(move_is_killer(move, sort->killer))
                    continue;

                if(!pseudo_is_legal(move, sort->board))
                    continue;
                sort->value = history_prob(move, sort->board, thread_id);
                }
            else
                {
                return 0;
                }
            return move;
            }
        // next stage
        gen = code[sort->gen++];

        if(false) { }
        else if(gen == GEN_TRANS)
            {
            LIST_CLEAR(sort->list);

            if(sort->trans_killer != 0)
                LIST_ADD(sort->list, sort->trans_killer);
            sort->test = TEST_TRANS_KILLER;
            }
        else if(gen == GEN_GOOD_CAPTURE)
            {
            gen_captures(sort->list, sort->board);
            sort->capture_nb = LIST_SIZE(sort->list);
            note_mvv_lva(sort->list, sort->board);
            list_sort(sort->list);
            LIST_CLEAR(sort->bad);
            sort->test = TEST_GOOD_CAPTURE;
            }
        else if(gen == GEN_BAD_CAPTURE)
            {
            list_copy(sort->list, sort->bad);
            sort->test = TEST_BAD_CAPTURE;
            }
        else if(gen == GEN_KILLER)
            {
            LIST_CLEAR(sort->list);

            for (i = 0; i < 2; i++)
                {
                if(sort->killer[i] != 0)
                    LIST_ADD(sort->list, sort->killer[i]);
                }
            sort->test = TEST_KILLER;
            }
        else if(gen == GEN_QUIET)
            {
            gen_quiet_moves(sort->list, sort->board);
            note_quiet_moves(sort->list, sort->board, thread_id);
            list_sort(sort->list);
            sort->test = TEST_QUIET;
            }
        else
            {
            return 0;
            }
        sort->pos = 0;
        }
    }

// sort_init_qs()
void sort_init_qs(sort_t *sort, board_t *board, const attack_t *attack, bool check)
    {
    sort->board = board;
    sort->attack = attack;

    if(ATTACK_IN_CHECK(sort->attack))
        {
        sort->gen = pos_evasionqs;
        }
    else if(check)
        {
        sort->gen = pos_check_qs;
        }
    else
        {
        sort->gen = pos_capture_qs;
        }
    LIST_CLEAR(sort->list);
    sort->pos = 0;
    }

// sort_next_qs()
int sort_next_qs(sort_t *sort)
    {
    int move;
    int gen;

    while(true)
        {
        while(sort->pos < LIST_SIZE(sort->list))
            {
            // next move
            move = LIST_MOVE(sort->list, sort->pos);
            sort->pos++;

            // test
            if(false) { }
            else if(sort->test == TEST_LEGAL)
                {
                if(!pseudo_is_legal(move, sort->board))
                    continue;
                }
            else if(sort->test == TEST_CAPTURE_QS)
                {
                if(!capture_is_good(move, sort->board))
                    continue;

                if(!pseudo_is_legal(move, sort->board))
                    continue;
                }
            else if(sort->test == TEST_CHECK_QS)
                {
                if(see_move(move, sort->board) < 0)
                    continue;

                if(!pseudo_is_legal(move, sort->board))
                    continue;
                }
            else
                {
                return 0;
                }
            return move;
            }
        // next stage
        gen = code[sort->gen++];

        if(false) { }
        else if(gen == GEN_EVASION_QS)
            {
            gen_pseudo_evasions(sort->list, sort->board, sort->attack);
            note_moves_simple(sort->list, sort->board);
            list_sort(sort->list);
            sort->test = TEST_LEGAL;
            }
        else if(gen == GEN_CAPTURE_QS)
            {
            gen_captures(sort->list, sort->board);
            note_mvv_lva(sort->list, sort->board);
            list_sort(sort->list);
            sort->test = TEST_CAPTURE_QS;
            }
        else if(gen == GEN_CHECK_QS)
            {
            gen_quiet_checks(sort->list, sort->board);
            sort->test = TEST_CHECK_QS;
            }
        else
            {
            return 0;
            }
        sort->pos = 0;
        }
    }

// good_move()
void good_move(int move, const board_t *board, int depth, int height, int thread_id)
    {
    uint16 index;
    int i;

    if(move_is_tactical(move, board))
        return;
    // killer
    store_killer(move, height, thread_id);
    // history
    index = history_index(move, board);
    history[thread_id][index] += HISTORY_INC(depth);

    if(history[thread_id][index] >= 16384)
        {
        for (i = 0; i < 768; i++)
            {
            history[thread_id][i] = (history[thread_id][i] + 1) / 2;
            }
        }
    }

// history_good()
void history_good(int move, const board_t *board, int thread_id)
    {
    uint16 index;

    if(move_is_tactical(move, board))
        return;

    // history
    index = history_index(move, board);
    hist_hit[thread_id][index]++;
    hist_tot[thread_id][index]++;

    if(hist_tot[thread_id][index] >= 16384)
        {
        hist_hit[thread_id][index] = (hist_hit[thread_id][index] + 1) / 2;
        hist_tot[thread_id][index] = (hist_tot[thread_id][index] + 1) / 2;
        }
    }

// history_bad()
void history_bad(int move, const board_t *board, int thread_id)
    {
    uint16 index;

    if(move_is_tactical(move, board))
        return;

    // history
    index = history_index(move, board);
    hist_tot[thread_id][index]++;

    if(hist_tot[thread_id][index] >= 16384)
        {
        hist_hit[thread_id][index] = (hist_hit[thread_id][index] + 1) / 2;
        hist_tot[thread_id][index] = (hist_tot[thread_id][index] + 1) / 2;
        }
    }

void history_reset(int move, const board_t *board, int thread_id)
    {
    uint16 index;
    // history
    index = history_index(move, board);
    hist_hit[thread_id][index] = 1;
    hist_tot[thread_id][index] = 1;
    }

// note_moves()
void note_moves(list_t *list, const board_t *board, int height, int trans_killer, int thread_id)
    {
    int size;
    int i, move;
    size = LIST_SIZE(list);

    if(size >= 2)
        {
        for (i = 0; i < size; i++)
            {
            move = LIST_MOVE(list, i);
            list->value[i] = move_value(move, board, height, trans_killer, thread_id);
            }
        }
    }

// note_quiet_moves()
static void note_quiet_moves(list_t *list, const board_t *board, int thread_id)
    {
    int size;
    int i, move;
    size = LIST_SIZE(list);

    if(size >= 2)
        {
        for (i = 0; i < size; i++)
            {
            move = LIST_MOVE(list, i);
            list->value[i] = quiet_move_value(move, board, thread_id);
            }
        }
    }

// note_moves_simple()
static void note_moves_simple(list_t *list, const board_t *board)
    {
    int size;
    int i, move;
    size = LIST_SIZE(list);

    if(size >= 2)
        {
        for (i = 0; i < size; i++)
            {
            move = LIST_MOVE(list, i);
            list->value[i] = move_value_simple(move, board);
            }
        }
    }

// note_mvv_lva()
static void note_mvv_lva(list_t *list, const board_t *board)
    {
    int size;
    int i, move;
    size = LIST_SIZE(list);

    if(size >= 2)
        {
        for (i = 0; i < size; i++)
            {
            move = LIST_MOVE(list, i);
            list->value[i] = mvv_lva(move, board);
            }
        }
    }

// move_value()
static int move_value(int move, const board_t *board, int height, int trans_killer, int thread_id)
    {
    int value, killer_nb;

    if(false) { }
    else if(move == trans_killer)
        { // transposition table killer
        value = 32766;
        }
    else if(move_is_tactical(move, board))
        { // capture or promote
        value = capture_value(move, board);
        }
    else if((killer_nb = killer_number(move, height, thread_id)) >= 0)
        { // killer
        value = 4 - killer_nb;
        }
    else
        { // quiet move
        value = quiet_move_value(move, board, thread_id);
        }
    return value;
    }

// capture_value()
static int capture_value(int move, const board_t *board)
    {
    int value;
    value = mvv_lva(move, board);

    if(capture_is_good(move, board))
        {
        value += 4000;
        }
    else
        {
        value += -28000;
        }
    return value;
    }

// quiet_move_value()
static int quiet_move_value(int move, const board_t *board, int thread_id)
    {
    int value;
    uint16 index;
    index = history_index(move, board);
    value = -24000 + history[thread_id][index];
    return value;
    }

// move_value_simple()
static int move_value_simple(int move, const board_t *board)
    {
    int value;
    value = -24000;

    if(move_is_tactical(move, board))
        value = mvv_lva(move, board);
    return value;
    }

// history_prob()
static int history_prob(int move, const board_t *board, int thread_id)
    {
    int value;
    uint16 index;
    index = history_index(move, board);
    value = (hist_hit[thread_id][index] * 16384) / hist_tot[thread_id][index];
    return value;
    }

// store_killer()
static void store_killer(int move, int height, int thread_id)
    {
    int i, j;

    for (i = 0; i < 2 && move != killer[thread_id][height][i]; i++);

    for (j = MIN(2 - 1, i); j > 0; j--)
        {
        killer[thread_id][height][j] = killer[thread_id][height][j - 1];
        }
    killer[thread_id][height][0] = move;
    }

// killer_number()
static int killer_number(int move, int height, int thread_id)
    {
    for (int i = 0; i < 2; i++)
        {
        if(move == killer[thread_id][height][i])
            return i;
        }
    return -1;
    }

// move_is_killer()
static bool move_is_killer(int move, int killer [])
    {
    for (int i = 0; i < 2; i++)
        if(move == killer[i])
            return true;
    return false;
    }

// capture_is_good()
static bool capture_is_good(int move, const board_t *board)
    {
    int piece, capture;

    // special cases
    if(MOVE_IS_EN_PASSANT(move))
        return true;

    if(move_is_under_promote(move))
        return false; // REMOVE ME?
    // captures and queen promotes
    capture = board->square[MOVE_TO(move)];

    if(capture != 0)
        {
        // capture
        if(MOVE_IS_PROMOTE(move))
            return true; // promote-capture
        piece = board->square[MOVE_FROM(move)];

        if(VALUE_PIECE(capture) >= VALUE_PIECE(piece))
            return true;
        }
    return see_move(move, board) >= 0;
    }

// mvv_lva()
static int mvv_lva(int move, const board_t *board)
    {
    int piece, capture, promote;
    int value;

    if(MOVE_IS_EN_PASSANT(move))
        {          // en-passant capture
        value = 5; // PxP
        }
    else if((capture = board->square[MOVE_TO(move)]) != 0)
        { // normal capture
        piece = board->square[MOVE_FROM(move)];
        value = PIECE_ORDER(capture) * 6 - PIECE_ORDER(piece) + 5;
        }
    else
        { // promote
        promote = move_promote(move);
        value = PIECE_ORDER(promote) - 5;
        }
    return value;
    }

// history_index()
static uint16 history_index(int move, const board_t *board)
    {
    uint16 index;
    index = PIECE_TO_12(board->square[MOVE_FROM(move)]) * 64 + SQUARE_TO_64(MOVE_TO(move));
    return index;
    }
// end of sort.cpp
